from typing import Callable

from pydantic.json_schema import SkipJsonSchema
from sqlalchemy import String, asc, cast, desc
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.sql.selectable import Select


def order_by(order_by: str | SkipJsonSchema[None] = None) -> Callable[[Select], Select]:
    def _order_by(query: Select) -> Select:
        if order_by:
            direction = desc if order_by.startswith("-") else asc
            column_name = order_by.lstrip("-")

            # Get the column from the query
            for column in query.inner_columns:
                if column.key == column_name:
                    # If the column is a UUID, cast it to a string
                    if isinstance(column.type, UUID):
                        column = cast(column, String)
                    query = query.order_by(direction(column))
                    break
        return query

    return _order_by



from .app import *  # noqa
from .crud import *  # noqa
from .db import *  # noqa
from .models import *  # noqa
from .pagination import *  # noqa
from .schemas import *  # noqa
from .testing import *  # noqa



import json

from httpx import ASGITransport, AsyncClient


class AsyncGenericClient:
    def __init__(self, app, headers={}):
        self.c = AsyncClient(
            transport=ASGITransport(app=app), base_url="http://testserver", follow_redirects=True
        )
        self.default_headers = headers

    async def get(self, url: str, r_code: int = 200, parse_json=True):
        re = await self.c.get(url, headers=self.default_headers)
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if parse_json else re.content

    async def delete(self, url: str, r_code: int = 204):
        re = await self.c.delete(url, headers=self.default_headers)
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if r_code != 204 else None

    async def post(
        self, url: str, obj: dict | str = {}, r_code: int = 201, raw_response=False, *args, **kwargs
    ):
        re = await self.c.post(
            url,
            data=json.dumps(obj) if isinstance(obj, dict) else obj,
            headers=self.default_headers | {"Content-Type": "application/json"},
            *args,
            **kwargs,
        )
        if re.status_code != r_code:
            print(re.content)
        if not raw_response:
            assert r_code == re.status_code
        return re.json() if not raw_response else re

    async def post_file(
        self, url: str, file, r_code: int = 201, raw_response=False, *args, **kwargs
    ):
        re = await self.c.post(
            url,
            files={"file": file},
            headers=self.default_headers,
            *args,
            **kwargs,
        )
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if not raw_response else re

    async def patch(
        self, url: str, obj: dict | str = {}, r_code: int = 200, raw_response=False, *args, **kwargs
    ):
        re = await self.c.patch(
            url,
            data=json.dumps(obj) if isinstance(obj, dict) else obj,
            headers=self.default_headers | {"Content-Type": "application/json"},
            *args,
            **kwargs,
        )
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if not raw_response else re

    async def put(
        self, url: str, obj: dict | str = {}, r_code: int = 200, raw_response=False, *args, **kwargs
    ):
        re = await self.c.put(
            url,
            data=json.dumps(obj) if isinstance(obj, dict) else obj,
            headers=self.default_headers
            | {
                "Content-Type": "application/json",
                "accept": "application/json",
            },
            *args,
            **kwargs,
        )
        if re.status_code != r_code:
            print(re.content)
        assert r_code == re.status_code
        return re.json() if not raw_response else re

    async def obj_lifecycle(
        self,
        input_obj: dict,
        url: str,
        pagination: bool = True,
        id_field: str = "id",
        created_at_check: bool = True,
    ):
        # GET LIST
        re = await self.get(url)
        if pagination:
            assert re["total"] == 0
            assert len(re["results"]) == 0
        else:
            assert len(re) == 0

        # CREATE
        re = await self.post(url, obj=input_obj)
        assert id_field in re
        assert re[id_field] is not None

        if created_at_check:
            assert "created_at" in re
            assert re["created_at"] is not None

        obj_id = str(re[id_field])

        # GET
        re = await self.get(f"{url}{obj_id}/")
        assert re[id_field] == obj_id

        # GET LIST
        re = await self.get(url)
        if pagination:
            assert re["total"] == 1
            assert len(re["results"]) == 1
        else:
            assert len(re) == 1

        # DELETE
        await self.delete(f"{url}{obj_id}")

        # GET LIST
        re = await self.get(url)
        if pagination:
            assert re["total"] == 0
            assert len(re["results"]) == 0
        else:
            assert len(re) == 0

        # GET
        await self.get(f"{url}{obj_id}", parse_json=False, r_code=404)



try:
    import sqlalchemy

    from .async_session import *
    from .helpers import *
    from .session import *
except ImportError:
    print("SQLAlchemy not installed. Database functionality will be disabled.")



from sqlalchemy_utils import create_database, database_exists


def create_if_not_exists(db_name: str):
    from .common import SQLALCHEMY_DATABASE_URL

    if not database_exists(SQLALCHEMY_DATABASE_URL + db_name):
        create_database(SQLALCHEMY_DATABASE_URL + db_name)



from typing import Optional

import jwt
import requests
from fastapi import HTTPException, Request, Security
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer

from creyPY.helpers import create_random_password

from .common import (
    AUTH0_ALGORIGHM,
    AUTH0_AUDIENCE,
    AUTH0_CLIENT_ID,
    AUTH0_DOMAIN,
    AUTH0_ISSUER,
)
from .exceptions import UnauthenticatedException, UnauthorizedException
from .manage import get_management_token

JWKS_CLIENT = jwt.PyJWKClient(f"https://{AUTH0_DOMAIN}/.well-known/jwks.json")


async def verify(
    request: Request,
    token: Optional[HTTPAuthorizationCredentials] = Security(HTTPBearer(auto_error=False)),
) -> str:
    if token is None:
        raise UnauthenticatedException

    # This gets the 'kid' from the passed token
    try:
        signing_key = JWKS_CLIENT.get_signing_key_from_jwt(token.credentials).key
    except jwt.exceptions.PyJWKClientError as error:
        raise UnauthorizedException(str(error))
    except jwt.exceptions.DecodeError as error:
        raise UnauthorizedException(str(error))

    try:
        payload = jwt.decode(
            token.credentials,
            signing_key,
            algorithms=[AUTH0_ALGORIGHM],
            audience=AUTH0_AUDIENCE,
            issuer=AUTH0_ISSUER,
        )
    except Exception as error:
        raise UnauthorizedException(str(error))

    return payload["sub"]


### GENERIC AUTH0 CALLS ###
def get_user(sub) -> dict:
    re = requests.get(
        f"https://{AUTH0_DOMAIN}/api/v2/users/{sub}",
        headers={"Authorization": f"Bearer {get_management_token()}"},
        timeout=5,
    )
    if re.status_code != 200:
        raise HTTPException(re.status_code, re.json())
    return re.json()


def patch_user(input_obj: dict, sub) -> dict:
    re = requests.patch(
        f"https://{AUTH0_DOMAIN}/api/v2/users/{sub}",
        headers={"Authorization": f"Bearer {get_management_token()}"},
        json=input_obj,
        timeout=5,
    )
    if re.status_code != 200:
        raise HTTPException(re.status_code, re.json())
    return re.json()


### USER METADATA CALLS ###
def get_user_metadata(sub) -> dict:
    try:
        return get_user(sub).get("user_metadata", {})
    except:
        return {}


def patch_user_metadata(input_obj: dict, sub) -> dict:
    return patch_user({"user_metadata": input_obj}, sub)


def clear_user_metadata(sub) -> dict:
    return patch_user({"user_metadata": {}}, sub)


def request_verification_mail(sub: str) -> None:
    re = requests.post(
        f"https://{AUTH0_DOMAIN}/api/v2/jobs/verification-email",
        headers={"Authorization": f"Bearer {get_management_token()}"},
        json={"user_id": sub},
        timeout=5,
    )
    if re.status_code != 201:
        raise HTTPException(re.status_code, re.json())
    return re.json()


def create_user_invite(email: str, company_id: str) -> dict:
    re = requests.post(
        f"https://{AUTH0_DOMAIN}/api/v2/users",
        headers={"Authorization": f"Bearer {get_management_token()}"},
        json={
            "email": email,
            "connection": "Username-Password-Authentication",
            "password": create_random_password(),
            "verify_email": False,
            "app_metadata": {"invitedToMyApp": True},
            "user_metadata": {"company_ids": [company_id]},
        },
        timeout=5,
    )
    if re.status_code != 201:
        raise HTTPException(re.status_code, re.json())
    return re.json()


def delete_user_invite(user_id: str) -> None:
    re = requests.delete(
        f"https://{AUTH0_DOMAIN}/api/v2/users/{user_id}",
        headers={"Authorization": f"Bearer {get_management_token()}"},
        timeout=5,
    )
    if re.status_code != 204:
        raise HTTPException(re.status_code, re.json())


def password_change_mail(email: str) -> bool:
    re = requests.post(
        f"https://{AUTH0_DOMAIN}/dbconnections/change_password",
        headers={"Authorization": f"Bearer {get_management_token()}"},
        json={
            "client_id": AUTH0_CLIENT_ID,
            "email": email,
            "connection": "Username-Password-Authentication",
        },
        timeout=5,
    )

    if re.status_code != 200:
        raise HTTPException(re.status_code, re.json())
    return True
